# slightly modified version of the code made available by Guosheng Fu:
# https://ngsolve.org/forum/ngspy-forum/114-interface-integrals#491

from ngsolve import *
from netgen.geom2d import SplineGeometry
from math import pi
ngsglobals.msg_level=1

# Parameters
order = 2
hmax = 1.0/32

nu = 1
kappainv = 1

dcFlag = False
ndofMax = 10

def topBottom():
    geometry = SplineGeometry()
    pnts     = [ (0, 0), (1, 0),  (1, 0.5), (1, 1), (0, 1), (0, 0.5)]
    pnums    = [geometry.AppendPoint(*p) for p in pnts]
    # start-point, end-point, boundary-condition, domain on left side, domain on right side:
    lines = [ (pnums[0], pnums[1], "gammaD",  2, 0),
              (pnums[1], pnums[2], "gammaD",  2, 0), 
              (pnums[2], pnums[3], "gammaS",  1, 0),
              (pnums[3], pnums[4], "gammaS",  1, 0),
              (pnums[4], pnums[5], "gammaS",  1, 0),
              (pnums[5], pnums[0], "gammaD",  2, 0),
              (pnums[5], pnums[2], "gammaSD", 1, 2)]
    for p1,p2,bc,left,right in lines:
        geometry.Append( ["line", p1, p2], bc=bc, leftdomain=left, rightdomain=right)

    geometry.SetMaterial(1, "stokesfluid")
    geometry.SetMaterial(2, "darcyfluid")
        
    return geometry

stokesId = [1, 0]
darcyId  = [0, 1]

mesh = Mesh(topBottom().GenerateMesh (maxh=hmax))

def GetInterfaceIndicator(mesh):
    dom_indicator = GridFunction( L2(mesh, order=0) )
    dom_indicator.Set( CoefficientFunction([1,2]) )
  
    ffes            = FacetFESpace(mesh, order=0)
    dom_ind_average = GridFunction(ffes)
    mf  = BilinearForm(ffes, symmetric=True)
    mf += SymbolicBFI(ffes.TrialFunction()*ffes.TestFunction(), element_boundary=True)
    mf.Assemble()
    ff  = LinearForm(ffes)
    ff += SymbolicLFI(dom_indicator*ffes.TestFunction(), element_boundary=True)
    ff.Assemble()
    dom_ind_average.vec.data = mf.mat.Inverse() * ff.vec

    interface_indicator = BitArray(ffes.ndof)
    for i in range(len(dom_ind_average.vec)):
        if (dom_ind_average.vec[i] > 1.75) or (dom_ind_average.vec[i] < 1.25):
            interface_indicator[i] = False
        else:
            interface_indicator[i] = True
    #print(sum(interface_indicator))
    return interface_indicator

n = specialcf.normal(mesh.dim)
h = specialcf.mesh_size

# Exact solution
us1 = -(1/(2*pi*pi))*sin(pi*x)*exp(y/2)
us2 = (1/pi)*cos(pi*x)*exp(y/2)
exact_us = CoefficientFunction((us1, us2))
exact_ps = CoefficientFunction( -(1/pi)*cos(pi*x)*exp(y/2) )

dus1_1 = -(1/(2*pi))*cos(pi*x)*exp(y/2)
dus1_2 = -(1/(4*pi*pi))*sin(pi*x)*exp(y/2)
dus2_1 = -sin(pi*x)*exp(y/2)
dus2_2 = (1/(2*pi))*cos(pi*x)*exp(y/2)

dus1_11 = (1/(2*pi))*pi*sin(pi*x)*exp(y/2)
dus1_22 = -(1/(4*pi*pi))*sin(pi*x)*(1/2)*exp(y/2)
dus2_12 = -sin(pi*x)*(1/2)*exp(y/2)
dus2_11 = -pi*cos(pi*x)*exp(y/2)
dus1_21 = -(1/(4*pi*pi))*pi*cos(pi*x)*exp(y/2)
dus2_22 = (1/(2*pi))*cos(pi*x)*(1/2)*exp(y/2)

ps_1 = sin(pi*x)*exp(y/2)
ps_2 = -(1/(2*pi))*cos(pi*x)*exp(y/2)

fs1 = -2.0*(dus1_11 + 0.5*dus1_22 + 0.5*dus2_12) + ps_1
fs2 = -2.0*(0.5*dus2_11 + 0.5*dus1_21 + dus2_22) + ps_2

source_s = CoefficientFunction((fs1, fs2))

ud1 = -2*sin(pi*x)*exp(y/2)
ud2 =  (1/pi)*cos(pi*x)*exp(y/2)
exact_ud = CoefficientFunction((ud1, ud2))
exact_pd = CoefficientFunction(-(2/pi)*cos(pi*x)*exp(y/2))

pd_1 = 2*sin(pi*x)*exp(y/2)
pd_2 = -(1/pi)*cos(pi*x)*exp(y/2)

ud1_1 = -2*pi*cos(pi*x)*exp(y/2)
ud2_2 = (1/pi)*cos(pi*x)*(1/2)*exp(y/2)

source_d = CoefficientFunction(-ud1_1 - ud2_2)

# Function spaces
V    = HDiv(mesh, order = order, dirichlet="gammaS|gammaD")
Vbar = FESpace("vectorfacet", mesh, order = order, dirichlet="gammaS", definedon=mesh.Materials("stokesfluid"))
Q    = L2(mesh, order = order - 1)
R    = FESpace("number", mesh)
X    = FESpace([V, Vbar, Q, R], dgjumps=False)
u, ubar, p, p0 = X.TrialFunction()
v, vbar, q, q0 = X.TestFunction()

def symgrad(u):
    return 0.5*(grad(u) + grad(u).trans)

def tangent(u):
    return u - (u*n)*n
  
interface_indicator = GetInterfaceIndicator(mesh)

dom1 = GridFunction(L2(mesh, order=0))
dom2 = GridFunction(L2(mesh, order=0))

# Penalty parameter
beta = 10*order*order

# Bi-linear form
a = BilinearForm(X)

# Stokes
stokesCell  = InnerProduct(2*nu*symgrad(u), symgrad(v)) \
              - div(u)*q - div(v)*p + p*q0 + q*p0
stokesFacet = (beta*nu/h)*InnerProduct(tangent(vbar-v), tangent(ubar-u)) \
              - 2*nu*InnerProduct ( symgrad(u)*n, tangent(v-vbar) ) \
              - 2*nu*InnerProduct ( symgrad(v)*n, tangent(u-ubar) )

# Darcy
darcyCell = kappainv*u*v - div(u)*q - div(v)*p + p*q0 + q*p0

# interface (only stokes part is needed)
interfaceTerm = 0.5*(1+4*pi*pi)*tangent(vbar)*tangent(ubar)

a += SymbolicBFI ( stokesCell, definedon=mesh.Materials("stokesfluid") )
a += SymbolicBFI ( darcyCell , definedon=mesh.Materials("darcyfluid") )
a += SymbolicBFI ( stokesFacet, element_boundary = True, definedon=mesh.Materials("stokesfluid") )
interfacebfi = SymbolicBFI (interfaceTerm, VOL, skeleton=True) 
interfacebfi.SetDefinedOnElements(interface_indicator)
a += interfacebfi

# Linear form
f = LinearForm(X)
f += SymbolicLFI(source_s*v, definedon=mesh.Materials("stokesfluid"))
f += SymbolicLFI(source_d*q, definedon=mesh.Materials("darcyfluid"))

# For solver
c = Preconditioner(type="direct", bf=a, inverse="pardiso")

# gridfunction for the solution
gfu = GridFunction(X)

def CalcL2Error(sol):
    err_us = sqrt(Integrate((sol.components[0]-exact_us)**2, mesh, order = 2*order, definedon=mesh.Materials("stokesfluid")))
    err_ud = sqrt(Integrate((sol.components[0]-exact_ud)**2, mesh, order = 2*order, definedon=mesh.Materials("darcyfluid")))
    return (err_us, err_ud)


store = []
def SolveBVP():

  dom1.Update()
  dom1.Set(CoefficientFunction(stokesId))
  dom2.Update()
  dom2.Set(CoefficientFunction(darcyId))

  X.Update()
  print("ndof:", X.ndof)  
  gfu.Update()
  
  interface_indicator = GetInterfaceIndicator(mesh)
  interfacebfi.SetDefinedOnElements(interface_indicator)
    
  # boundary condition
  gfu.components[0].Set(dom1*exact_us+dom2*exact_ud)
  gfu.components[1].Set(exact_us, definedon=mesh.Boundaries("gammaS"))

  a.Assemble()
  f.Assemble()

  BVP(bf=a, lf=f, gf=gfu, pre=c, maxsteps=10, prec=1.0e-16).Do()
  err_us, err_ud = CalcL2Error(gfu)
  print(err_us, err_ud)

  velocity = CoefficientFunction (gfu.components[0])
  Draw(velocity, mesh, "velocity", sd=3)

  pressure = CoefficientFunction (gfu.components[2])
  Draw (pressure, mesh, "pressure", sd=3)
    
SolveBVP()
